@@ -168,13 +168,18 @@ class ViewAdapter(swtsimulator.WidgetAdapter):
         return "View"
 
 class PartActivateEvent(storytext.guishared.GuiEvent):
+    allInstances = []
+    def __init__(self, *args, **kw):
+        storytext.guishared.GuiEvent.__init__(self, *args, **kw)
+        self.allInstances.append(self)
+        
     def connectRecord(self, method):
         class RecordListener(IPartListener):
             def partActivated(listenerSelf, part):#@NoSelf
                 storytext.guishared.catchAll(method, part, self)
         page = self.widget.getViewReference().getPage()
         swtsimulator.runOnUIThread(page.addPartListener, RecordListener())
-
+        
     def generate(self, *args):
         # The idea is to just do this, but it seems to cause strange things to happen
         #internally. So we do it outside SWTBot instead.
@@ -187,28 +192,52 @@ class PartActivateEvent(storytext.guishared.GuiEvent):
         # The idea is to just do this, but it seems to cause strange things to happen
         #internally. So we do it outside SWTBot instead.
         #self.widget.setFocus()
-        if self.widget.getViewReference().getTitle() == argumentString:
+        if self.getTitleArgument() == argumentString:
             return argumentString
         else:
             raise UseCaseScriptError, "Could not find View named " + repr(argumentString) + " to activate."
 
+    def getTitle(self):
+        return self.widget.getViewReference().getTitle()
+
     def shouldRecord(self, part, *args):
         # TODO: Need to check no other events are waiting in DisplayFilter 
-        return self.widget.getViewReference().getTitle() == part.getTitle() and not swtsimulator.DisplayFilter.instance.hasEvents()
+        return part is self.widget.getViewReference().getView(False) and self.hasMultipleViews() and not swtsimulator.DisplayFilter.instance.hasEvents()
+
+    def hasMultipleViews(self):
+        # If there is only one view, don't try to record if it's activated, it's probably just been created...
+        return sum((i.isActivatable() for i in self.allInstances)) > 1
+
+    def isActivatable(self):
+        return self.getControl() is not None
 
     def delayLevel(self, part, *args):
         # If there are events for other shells, implies we should delay as we're in a dialog
         return swtsimulator.DisplayFilter.instance.otherEventCount(part, [])
     
+    def getControl(self):
+        return self.widget.getViewReference().getPane().getControl()
+    
+    def getTitleArgument(self):
+        # Handle multiple parts with the same title...
+        title = self.getTitle()
+        index = 1
+        for otherInstance in self.allInstances:
+            if otherInstance is self:
+                break
+            if otherInstance.getTitle() == title and otherInstance.getControl() is not None:
+                index += 1
+                
+        if index > 1:
+            title += " (" + str(index) + ")"
+        return title
+    
     def outputForScript(self, *args):
-        return ' '.join([self.name, self.widget.getViewReference().getTitle() ])
+        return ' '.join([self.name, self.getTitleArgument() ])
 
     def isStateChange(self):
         return True
-    
-    def isImpliedByCTabClose(self, tab):
-        return True
-    
+        
     def implies(self, stateChangeOutput, stateChangeEvent, *args):
         return isinstance(stateChangeEvent, PartActivateEvent)
     
@@ -219,7 +248,64 @@ class PartActivateEvent(storytext.guishared.GuiEvent):
     @classmethod
     def getAssociatedSignal(cls, widget):
         return "ActivatePart"
+    
+class PartCloseEvent(storytext.guishared.GuiEvent):
+    def connectRecord(self, method):
+        class RecordListener(IPartListener):
+            def partClosed(listenerSelf, part): #@NoSelf
+                if part is self.widget.getViewReference().getView(False):
+                    storytext.guishared.catchAll(method, part, self)
+        page = self.widget.getViewReference().getPage()
+        swtsimulator.runOnUIThread(page.addPartListener, RecordListener())
+        
+    def shouldRecord(self, *args):
+        return False # Only point of this is to prevent false Activate Events from being recorded
+        
+    def implies(self, stateChangeOutput, stateChangeEvent, *args):
+        return isinstance(stateChangeEvent, (PartActivateEvent, swtsimulator.CTabSelectEvent))
+    
+    def checkPreviousWhenRejected(self):
+        return True
+    
+    @classmethod
+    def getSignalsToFilter(cls):
+        return []
+    
+    @classmethod
+    def getAssociatedSignal(cls, widget):
+        return "ClosePart"
+    
+class RCPCTabSelectEvent(swtsimulator.CTabSelectEvent):
+    def implies(self, stateChangeOutput, stateChangeEvent, *args):
+        return isinstance(stateChangeEvent, PartActivateEvent) or swtsimulator.CTabSelectEvent.implies(self, stateChangeOutput, stateChangeEvent, *args)
 
+    
+# Removing the last tab in a view folder causes the entire folder to be disposed
+# To record the correct things we need to cache their names before they are disposed.
+class RCPCTabCloseEvent(swtsimulator.CTabCloseEvent):
+    disposedTabs = {}
+    disposeFilter = None
+    def connectRecord(self, method):
+        swtsimulator.CTabCloseEvent.connectRecord(self, method)
+        class DisposeFilter(Listener):
+            def handleEvent(listenerSelf, e): #@NoSelf
+                if isinstance(e.widget, CTabItem):
+                    self.disposedTabs[e.widget] = e.widget.getText()
+                    
+        if not self.disposeFilter:
+            RCPCTabCloseEvent.disposeFilter = DisposeFilter()
+            swtsimulator.runOnUIThread(self.widget.widget.widget.getDisplay().addFilter, SWT.Dispose, self.disposeFilter)
+
+    def implies(self, stateChangeOutput, stateChangeEvent, *args):
+        return isinstance(stateChangeEvent, PartActivateEvent) or swtsimulator.CTabCloseEvent.implies(self, stateChangeOutput, stateChangeEvent, *args)
+    
+    def getItemText(self, item):
+        if item.isDisposed():
+            return self.disposedTabs.get(item)
+        else:
+            return item.getText()
+
+    
 class SWTBotExpandableComposite(AbstractSWTBotControl):
     def clickOnCenter(self):
         firstChild = self.widget.getChildren()[0]
@@ -255,6 +341,13 @@ class ExpandableCompositeEvent(swtsimulator.SelectEvent):
 
         swtsimulator.runOnUIThread(self.widget.widget.widget.addExpansionListener, RecordListener())
             
-swtsimulator.eventTypes.append((swtbot.widgets.SWTBotView, [ PartActivateEvent ]))
+swtsimulator.eventTypes.append((swtbot.widgets.SWTBotView, [ PartActivateEvent, PartCloseEvent ]))
 swtsimulator.eventTypes.append((SWTBotExpandableComposite, [ ExpandableCompositeEvent ]))
 
+replacements = [(swtsimulator.CTabCloseEvent, RCPCTabCloseEvent),
+                (swtsimulator.CTabSelectEvent, RCPCTabSelectEvent)]
+
+for swtbotClass, eventClasses in swtsimulator.eventTypes:
+    for oldClass, newClass in replacements:
+        if oldClass in eventClasses:
+            eventClasses[eventClasses.index(oldClass)] = newClass
