@@ -21,17 +21,18 @@ class Describer(storytext.guishared.Describer):
     ignoreWidgets = [ types.NoneType ]
     # DateTime children are an implementation detail
     # Coolbars and Expandbars describe their children directly : they have two parallel children structures
-    ignoreChildren = (swt.widgets.CoolBar, swt.widgets.ExpandBar, swt.widgets.DateTime)
+    ignoreChildren = (swt.widgets.CoolBar, swt.widgets.ExpandBar, swt.widgets.DateTime, swt.widgets.Group)
     statelessWidgets = [ swt.widgets.Sash ]
     stateWidgets = [ swt.widgets.Shell, swt.widgets.Button, swt.widgets.Menu, swt.widgets.Link,
                      swt.widgets.CoolBar, swt.widgets.ToolBar, swt.widgets.Label, swt.custom.CLabel,
                      swt.widgets.Combo, swt.widgets.ExpandBar, swt.widgets.Text, swt.widgets.List,
                      swt.widgets.Tree, swt.widgets.DateTime, swt.widgets.TabFolder, swt.widgets.Table, 
                      swt.custom.CTabFolder, swt.widgets.Canvas, swt.browser.Browser, swt.custom.CCombo,
-                     swt.widgets.Composite ]
+                     swt.widgets.Spinner, swt.widgets.Group, swt.widgets.Composite ]
     childrenMethodName = "getChildren"
     visibleMethodName = "getVisible"
     def __init__(self):
+        storytext.guishared.Describer.__init__(self)
         self.canvasCounter = storytext.guishared.WidgetCounter()
         self.contextMenuCounter = storytext.guishared.WidgetCounter(self.contextMenusEqual)
         self.widgetsAppeared = []
@@ -41,7 +42,6 @@ class Describer(storytext.guishared.Describer):
         self.storedImages = {}
         self.imageToName = {}
         self.handleImages()
-        storytext.guishared.Describer.__init__(self)
 
     def setWidgetPainted(self, widget):
         if widget not in self.widgetsDescribed and widget not in self.windows and widget not in self.widgetsAppeared:
@@ -66,13 +66,13 @@ class Describer(storytext.guishared.Describer):
 
         display.addFilter(swt.SWT.Show, ShowListener())
         display.addFilter(swt.SWT.Paint, PaintListener())
-
+        
     def describeWithUpdates(self, shellMethod):
         shell = shellMethod()
         if shell in self.windows:
             stateChanges = self.findStateChanges(shell)
             stateChangeWidgets = [ widget for widget, old, new in stateChanges ]
-            if self.structureLogger.isEnabledFor(logging.DEBUG):
+            if self.structureLog.isEnabledFor(logging.DEBUG):
                 for widget in stateChangeWidgets:
                     self.describeStructure(widget)
             describedForAppearance = self.describeAppearedWidgets(stateChangeWidgets)
@@ -83,11 +83,8 @@ class Describer(storytext.guishared.Describer):
         self.widgetsAppeared = []
         
     def shouldCheckForUpdates(self, widget, shell):
-        return not isinstance(widget, self.getMenuClasses()) or (not widget.isDisposed() and widget.getShell() == shell)
+        return not widget.isDisposed() and widget.getShell() == shell
     
-    def getMenuClasses(self):
-        return swt.widgets.Menu
-
     def widgetShowing(self, widget):
         return not widget.isDisposed() and util.isVisible(widget)
 
@@ -123,7 +120,7 @@ class Describer(storytext.guishared.Describer):
         descs = []
         for item in itemBar.getItems():
             currPrefix = prefix + " " * indent * 2
-            selected = item in selection
+            selected = item in selection or (hasattr(item, "getSelection") and item.getSelection())
             if columnCount:
                 row = [ self.getItemColumnDescription(item, i, currPrefix, selected) for i in range(columnCount) ]
                 descs.append(row)
@@ -260,6 +257,7 @@ class Describer(storytext.guishared.Describer):
     def storeImageData(self, url):
         imgDesc = ImageDescriptor.createFromURL(url)
         name = url.getFile()
+        self.logger.debug("Storing image data for file " + name)
         if imgDesc is not None:
             newImage = imgDesc.createImage()
             data = newImage.getImageData()
@@ -351,6 +349,8 @@ class Describer(storytext.guishared.Describer):
         decoText = self.getControlDecorationDescription(item)
         if decoText:
             elements.append(decoText)
+        if isinstance(item, swt.widgets.Spinner):
+            elements += self.getSpinnerPropertyElements(item)
         if hasattr(item, "getToolTipText") and item.getToolTipText():
             elements.append("Tooltip '" + item.getToolTipText() + "'")
         elements += self.getStyleDescriptions(item)
@@ -365,6 +365,20 @@ class Describer(storytext.guishared.Describer):
             elements.append("+")
         return elements
 
+    def getSpinnerPropertyElements(self, item):
+        elements = []
+        min = item.getMinimum()
+        if min != 0:
+            elements.append("Min " + str(min))
+        elements.append("Max " + str(item.getMaximum()))
+        step = item.getIncrement()
+        if step != 1:
+            elements.append("Step " + str(step))
+        step = item.getPageIncrement()
+        if step != 10:
+            elements.append("Page Step " + str(step))
+        return elements
+
     def getLabelState(self, label):
         if label.getStyle() & swt.SWT.SEPARATOR:
             if label.getStyle() & swt.SWT.VERTICAL:
@@ -421,18 +435,20 @@ class Describer(storytext.guishared.Describer):
 
     def getBrowserState(self, widget):
         url = util.getRealUrl(widget)
+        if url and not url.startswith("file://"):
+            return url
         # Ignore non-breaking spaces, they are invisible anyway
         # Webkit returns them in invalid format without the semicolons... handle that too.
         rawText = widget.getText().replace(u"\xa0", " ").replace("&nbsp;", " ").replace("&nbsp", " ")
         text = storytext.guishared.encodeToLocale(rawText)
-        return url or BrowserHtmlParser().parse(text)
+        return BrowserHtmlParser().parse(text)
         
     def getUpdatePrefix(self, widget, oldState, state):
-        if isinstance(widget, (self.getTextEntryClass(), swt.browser.Browser)):
-            return "\nUpdated " + (util.getTextLabel(widget) or "Text") +  " Field\n"
+        if isinstance(widget, (self.getTextEntryClass(), swt.browser.Browser, swt.widgets.Spinner)):
+            return "\nUpdated " + (util.getTextLabel(widget, useContext=True) or "Text") +  " Field\n"
         elif isinstance(widget, swt.widgets.Combo):
-            return "\nUpdated " + util.getTextLabel(widget) + " Combo Box\n"
-        elif util.getTopControl(widget):
+            return "\nUpdated " + util.getTextLabel(widget, useContext=True) + " Combo Box\n"
+        elif util.getTopControl(widget) or isinstance(widget, swt.widgets.Group):
             return "\n"
         elif isinstance(widget, swt.widgets.Menu):
             parentItem = widget.getParentItem()
@@ -479,6 +495,12 @@ class Describer(storytext.guishared.Describer):
 
     def getDateTimeDescription(self, widget):
         return self.getAndStoreState(widget)
+    
+    def getSpinnerDescription(self, widget):
+        return self.getTextDescription(widget)
+    
+    def getSpinnerState(self, widget):
+        return self.getTextState(widget)
 
     def getDateString(self, widget):
         if widget.getStyle() & swt.SWT.TIME:
@@ -567,11 +589,22 @@ class Describer(storytext.guishared.Describer):
 
     def getCompositeDescription(self, widget):
         return self.combineElements([ self.getStateControlDescription(widget), self.getContextMenuReference(widget) ])
+    
+    def getGroupState(self, widget):
+        return self.getCompositeState(widget), widget.getText()
+    
+    def getGroupDescription(self, widget):
+        header = "." * 6 + " " + widget.getText() + " " + "." * 6
+        footer = "." * len(header)
+        compositeDesc = self.getCompositeDescription(widget) or self.formatChildrenDescription(widget)
+        return header + "\n" + compositeDesc + "\n" + footer
 
     def getStateControlDescription(self, widget):
-        stateControl = self.getState(widget)
+        stateControlInfo = self.getState(widget)
+        stateControl = stateControlInfo[0] if isinstance(stateControlInfo, tuple) else stateControlInfo
+        if stateControlInfo:
+            self.widgetsWithState[widget] = stateControlInfo
         if stateControl:
-            self.widgetsWithState[widget] = stateControl
             if len(widget.getChildren()) > 1:
                 header = "+" * 6 + " Stacked Layout " + "+" * 6
                 footer = "+" * len(header)
@@ -596,7 +629,7 @@ class Describer(storytext.guishared.Describer):
     def _getDescription(self, widget):
         self.widgetsDescribed.add(widget)
         desc = storytext.guishared.Describer._getDescription(self, widget)
-        if desc and isinstance(widget, (swt.widgets.ExpandBar, swt.widgets.Tree, swt.widgets.List)):
+        if desc and isinstance(widget, (swt.widgets.ExpandBar, swt.widgets.Tree, swt.widgets.List, swt.widgets.Table)):
             desc = str(desc) + self.formatContextMenuDescriptions()
         return desc
 
@@ -630,9 +663,10 @@ class Describer(storytext.guishared.Describer):
         text = ""
         if "Menu" not in self.excludeClassNames:
             for contextMenu, menuId in self.contextMenuCounter.getWidgetsForDescribe():
-                menuDesc = self.getMenuDescription(contextMenu)
-                text += "\n\nContext Menu " + str(menuId) + ":\n" + menuDesc
-                self.widgetsWithState[contextMenu] = self.getMenuState(contextMenu) 
+                if not contextMenu.isDisposed():
+                    menuDesc = self.getMenuDescription(contextMenu)
+                    text += "\n\nContext Menu " + str(menuId) + ":\n" + menuDesc
+                    self.widgetsWithState[contextMenu] = self.getMenuState(contextMenu) 
         return text
 
     def getHorizontalSpan(self, widget, columns):
@@ -645,13 +679,24 @@ class Describer(storytext.guishared.Describer):
     def usesGrid(self, widget):
         return isinstance(widget.getLayout(), swt.layout.GridLayout)
 
-    def getLayoutColumns(self, widget, childCount, *args):
+    def getLayoutColumns(self, widget, childCount, sortedChildren):
         layout = widget.getLayout()
         if hasattr(layout, "numColumns"):
             return layout.numColumns
         elif hasattr(layout, "type"):
             if layout.type == swt.SWT.HORIZONTAL:
                 return childCount
+        elif isinstance(layout, swt.layout.FormLayout):
+            currColumns, maxColumns = 1, 1
+            for child in sortedChildren:
+                layoutData = child.getLayoutData()
+                if layoutData.right and layoutData.right.control:
+                    currColumns += 1
+                    if currColumns > maxColumns:
+                        maxColumns = currColumns
+                else:
+                    currColumns = 1
+            return maxColumns
         return 1
 
     def getRawDataLayoutDetails(self, layout, *args):
