@@ -3,7 +3,7 @@
 
 import os, sys, signal, logging
 from copy import copy
-from replayer import ReplayScript
+import replayer
 from definitions import *
 
 try:
@@ -18,20 +18,74 @@ class RecordScript:
         self.fileForAppend = None
         self.shortcutTrackers = []
     
+    def findCompletedTracker(self, line):
+        bestTracker = None
+        for tracker in self.shortcutTrackers:
+            if tracker.updateCompletes(line) and (bestTracker is None or tracker.isLongerThan(bestTracker)):
+                bestTracker = tracker
+        
+        return bestTracker
+
+    def recordWithTracker(self, line, bestTracker):
+        for tracker in self.shortcutTrackers:
+            tracker.addCommand(line)
+        
+        if bestTracker:
+            newCommands = bestTracker.getNewCommands()
+            self.rerecord(newCommands)
+            for tracker in self.shortcutTrackers:
+                if tracker is not bestTracker:
+                    tracker.rerecord(newCommands)
+        else:
+            self._record(line)
+            
+    def findPartLineCompleting(self, eventPairs):
+        for eventPair in eventPairs:
+            partLine1, partLine2 = eventPair
+            partTracker = self.findCompletedTracker(partLine1)
+            if partTracker:
+                return partLine1, partTracker, partLine2
+            partTracker = self.findCompletedTracker(partLine2)
+            if partTracker:
+                return partLine2, partTracker, partLine1
+            
+        return None, None, None
+    
+    def isSplittable(self, line):
+        return line.startswith(waitCommandName) and ("," in line or "*" in line)
+    
+    @staticmethod
+    def splitLine(line):
+        eventPairs = set()
+        parsedEvents = replayer.parseWaitCommand(line)
+        for i, (baseEvent, count) in enumerate(parsedEvents):
+            otherEvents = parsedEvents[:i] + parsedEvents[i+1:]
+            if len(otherEvents):
+                basicPair = frozenset([ replayer.assembleWaitCommand(parsedEvents[i:i+1]),
+                              replayer.assembleWaitCommand(otherEvents) ])
+                eventPairs.add(basicPair)
+            if count > 1:
+                for j in range(1, count):
+                    currEvent = [ (baseEvent, j) ]
+                    currOthers = otherEvents + [ (baseEvent, count - j)]
+                    parts = (replayer.assembleWaitCommand(currEvent), replayer.assembleWaitCommand(currOthers))
+                    partSet = frozenset(parts)
+                    eventPairs.add(partSet if len(partSet) == 2 else parts)
+        return eventPairs
+
     def record(self, line):
         try:
-            self._record(line)
-            bestTracker = None
-            for tracker in self.shortcutTrackers:
-                if tracker.updateCompletes(line) and \
-                    (bestTracker is None or tracker.isLongerThan(bestTracker)):
-                    bestTracker = tracker
-            if bestTracker:
-                newCommands = bestTracker.getNewCommands()
-                self.rerecord(newCommands)
-                for tracker in self.shortcutTrackers:
-                    if tracker is not bestTracker:
-                        tracker.rerecord(newCommands)
+            bestTracker = self.findCompletedTracker(line)
+            if bestTracker is None and self.isSplittable(line):
+                eventPairs = self.splitLine(line)
+                partLine, partTracker, otherPartLine = self.findPartLineCompleting(eventPairs)
+                if partLine is not None:
+                    self.recordWithTracker(partLine, partTracker)
+                    self.recordWithTracker(otherPartLine, None)
+                    return
+                    
+            self.recordWithTracker(line, bestTracker)
+            
         except IOError:
             sys.stderr.write("ERROR: Unable to record " + repr(line) + " to file " + repr(self.scriptName) + "\n") 
     
@@ -71,7 +125,7 @@ class ShortcutTracker:
         self.reset()
 
     def reset(self):
-        self.replayScript = ReplayScript(self.replayScript.name, ignoreComments=True)
+        self.replayScript = replayer.ReplayScript(self.replayScript.name, ignoreComments=True)
         self.commandsForMatch = copy(self.commandsForMismatch)
         self.argsUsed = []
         self.currRegexp = self.replayScript.getCommandRegexp()
@@ -81,9 +135,16 @@ class ShortcutTracker:
 
     def updateCompletes(self, line):
         if self.currRegexp is None:
-            self.logger.debug("Ignore " +  self.replayScript.getShortcutName())
             return False # We already reached the end and should forever be ignored...
         match = self.currRegexp.match(line)
+        self.logger.debug("Update completes? " +  self.replayScript.getShortcutName() + ", " + self.currRegexp.pattern \
+                          + ", " +  repr(line) + ", " + repr(self.commandsForMismatch) + ", " + repr(self.commandsForMatch))
+        return match and self.replayScript.hasTerminated()
+    
+    def addCommand(self, line):
+        if self.currRegexp is None:
+            self.logger.debug("Ignore " +  self.replayScript.getShortcutName())  # We already reached the end and should forever be ignored...
+        match = self.currRegexp.match(line)
         if match:
             self.commandsForMismatch.append(line)
             self.logger.debug("Match " +  self.replayScript.getShortcutName() + ", " + self.currRegexp.pattern \
@@ -98,17 +159,15 @@ class ShortcutTracker:
                     self.argsUsed[argPos] = val
             else:
                 self.argsUsed += match.groups()
-            return not self.currRegexp
         else:
             if self.hasStarted():
                 self.reset()
                 self.logger.debug("Reset " + self.replayScript.getShortcutName() + ", " + repr(self.commandsForMismatch))
-                return self.updateCompletes(line)
+                self.addCommand(line)
             else:
                 self.commandsForMismatch.append(line)
                 self.reset()
-                return False
-        
+            
     def rerecord(self, newCommands):
         # Some other tracker has completed, include it in our unmatched commands...
         started = self.hasStarted()
@@ -132,7 +191,7 @@ class ShortcutTracker:
 
 class UseCaseRecorder:
     def __init__(self):
-        self.logger = logging.getLogger("usecase record")
+        self.logger = logging.getLogger("storytext record")
         # Store events we don't record at the top level, usually controls on recording...
         self.eventsBlockedTopLevel = []
         self.scripts = []
@@ -245,9 +304,10 @@ class UseCaseRecorder:
             if event.checkPreviousWhenRejected():
                 delayLevel = event.delayLevel(*args)
                 if delayLevel in self.stateChangeEventInfo:
-                    stateChangeOutput, stateChangeEvent, _ = self.stateChangeEventInfo[delayLevel]
+                    stateChangeOutput, stateChangeEvent, stateChangeAppEvents = self.stateChangeEventInfo[delayLevel]
                     if event.implies(stateChangeOutput, stateChangeEvent, *args):
                         self.logger.debug("Discarded event implies previous state change event, ignoring previous also")
+                        self.applicationEvents.update(stateChangeAppEvents)
                         del self.stateChangeEventInfo[delayLevel]
             return
         
@@ -327,15 +387,19 @@ class UseCaseRecorder:
                             nextLevelEvents = []
                 else:
                     nextLevelEvents.append((scriptOutput, delayLevel, source))
+
+    def popComment(self):
+        # Make this atomic, it can be called from different threads
+        try:
+            return self.comments.pop(0)
+        except IndexError:
+            pass
         
     def recordComments(self):
-        while len(self.comments) > 0:
-            comment = self.comments.pop(0)
-            if comment is not None:
-                self._record(comment)
-            else:
-                self.logger.debug("Got None in comment list, not recording more comments now")
-                break
+        comment = self.popComment()
+        while comment is not None:
+            self._record(comment)
+            comment = self.popComment()
                 
     def record(self, line, event=None):
         self.recordComments()
@@ -410,13 +474,6 @@ class UseCaseRecorder:
             return text
         else:
             return text + " * " + str(count)
-        
-    def parseMultiples(self, text):
-        words = text.split()
-        if len(words) > 2 and words[-2] == "*" and words[-1].isdigit():
-            return " ".join(words[:-2]), int(words[-1])
-        else:
-            return text, 1
 
     def reduceApplicationEventCount(self, name, categoryName, delayLevel, remainder):
         newEventName = self.makeMultiple(name, remainder)
@@ -449,7 +506,7 @@ class UseCaseRecorder:
         for appEventKey, eventName in self.applicationEvents.items():
             categoryName, delayLevel = appEventKey
             if matchFunction(eventName, delayLevel):
-                basicName, count = self.parseMultiples(eventName)
+                basicName, count = replayer.parseMultiples(eventName)
                 if count == 1:
                     self.logger.debug("Unregistering application event " + repr(eventName) + " in category " + repr(categoryName))
                     del self.applicationEvents[appEventKey]
